import enum
import pathlib

import pytest
import requests
import requests_mock
import typer.testing
from typing_extensions import Iterable, Iterator, NamedTuple

import ignoro


def assert_in_string(fragments: Iterable[str], string: str):
    __tracebackhide__ = True
    for fragment in fragments:
        assert fragment.lower() in string.lower()


class TestConsole(NamedTuple):
    __test__ = False
    runner: typer.testing.CliRunner
    cwd: pathlib.Path


class TemplateMock(NamedTuple):
    name: str
    header: str
    body: str
    content: str
    response: str


class MockErrors(str, enum.Enum):
    NOT_EXIST = "not-exist-error"
    NOT_FOUND = "not-found-error"
    TIMEOUT = "timeout-error"
    CONNECTION = "connection-error"


@pytest.fixture(scope="function")
def test_console(tmp_path: pathlib.Path) -> Iterator[TestConsole]:
    runner = typer.testing.CliRunner(mix_stderr=False)
    with runner.isolated_filesystem(tmp_path) as cwd:
        yield TestConsole(runner, pathlib.Path(cwd))


@pytest.fixture(scope="function", autouse=True)
def mock_requests(
    requests_mock: requests_mock.Mocker,
    template_list_names_mock: tuple[str, ...],
    foo_template_mock: TemplateMock,
    bar_template_mock: TemplateMock,
) -> None:
    requests_mock.get(f"{ignoro.BASE_URL}/list?format=lines", text="\n".join(template_list_names_mock))
    requests_mock.get(f"{ignoro.BASE_URL}/foo", text=foo_template_mock.response)
    requests_mock.get(f"{ignoro.BASE_URL}/bar", text=bar_template_mock.response)
    requests_mock.get(f"{ignoro.BASE_URL}/{MockErrors.NOT_FOUND.value}", status_code=404)
    requests_mock.get(f"{ignoro.BASE_URL}/{MockErrors.TIMEOUT.value}", exc=requests.exceptions.Timeout())
    requests_mock.get(f"{ignoro.BASE_URL}/{MockErrors.CONNECTION.value}", exc=requests.exceptions.ConnectionError)


@pytest.fixture(scope="function")
def template_list() -> ignoro.TemplateList:
    return ignoro.TemplateList()


@pytest.fixture(scope="session")
def foo_template_mock(foo_template_content_mock: str) -> TemplateMock:
    name = "foo"
    header = ignoro.Template._create_header(name)
    body = foo_template_content_mock
    content = f"{header}\n{body}"
    response = api_response_mock(name, body)
    return TemplateMock(name, header, body, content, response)


@pytest.fixture(scope="session")
def bar_template_mock(bar_template_content_mock: str) -> TemplateMock:
    name = "bar"
    header = ignoro.Template._create_header(name)
    body = bar_template_content_mock
    content = f"{header}\n{body}"
    response = api_response_mock(name, body)
    return TemplateMock(name, header, body, content, response)


@pytest.fixture(scope="session")
def template_list_names_mock() -> tuple[str, ...]:
    return (
        "bar",
        "dotdot",
        "double-dash",
        "foo",
        "foobar",
        "hoppy",
    )


@pytest.fixture(scope="session")
def foo_template_content_mock() -> str:
    return """# Used by dotenv library to load environment variables.
.env

# Ignore compiled files
*.fooc

# Ignore virtual environment folder
env/

# Ignore log files
*.log

# Ignore database files
*.db

# Ignore cache files
*.cache

# Ignore sensitive information
secrets.txt
"""


@pytest.fixture(scope="session")
def bar_template_content_mock() -> str:
    return """# Ignore bundler config.
/.bundle

# Ignore all logfiles and tempfiles.
/log/*
/tmp/*

# Ignore the default SQLite database.
/db/*.sqlite3

# Ignore all .env files.
.env*
"""


def api_response_mock(name: str, body: str) -> str:
    return f"""# Created by https://www.toptal.com/developers/gitignore/api/{name.lower()}
# Edit at https://www.toptal.com/developers/gitignore?templates={name.lower()}

### {name.capitalize()} ###
{body}

# End of https://www.toptal.com/developers/gitignore/api/{name.lower()}
"""



from __future__ import annotations

import collections
import collections.abc
import pathlib
import re

import requests
from typing_extensions import Callable, Iterable, Iterator, NamedTuple, Optional, Sequence, SupportsIndex

import ignoro

__all__ = ["Template", "TemplateList", "Gitignore"]


class _MetadataMatch(NamedTuple):
    """A match for a metadata line."""

    index: int
    match: re.Match[str]


class _FindMetadataMixin:
    """Mixin class to find metadata in a .gitignore file."""

    @staticmethod
    def _find_metadata(lines: Iterable[str], pattern: re.Pattern[str]) -> list[_MetadataMatch]:
        """Find metadata in a list of lines. Returns a named tuple containing the index and match object."""
        results = []
        for index, line in enumerate(lines):
            if match := pattern.search(line):
                results.append(_MetadataMatch(index, match))

        return results


class Template(_FindMetadataMixin):
    """A .gitignore template from gitignore.io."""

    def __init__(self, name: str, body: Optional[str] = None) -> None:
        """Initialize a .gitignore template from a name and body."""
        self.name = name.lower()
        self.header = self._create_header(self.name)
        self._body = body

    @property
    def body(self) -> str:
        if self._body is None:
            url = f"{ignoro.BASE_URL}/{self.name.lower()}"

            try:
                response = requests.get(url)
            except requests.exceptions.ConnectionError as err:
                raise ignoro.exceptions.ApiError(f"Failed to connect to '{url}'") from err
            except requests.exceptions.Timeout as err:
                raise ignoro.exceptions.ApiError(f"Connection to '{url}' timed out") from err

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as err:
                raise ignoro.exceptions.ApiError(f"Failed to fetch '{url}' because {err.response.reason}") from err

            self._body = self._extract_body(response.text)

        return self._body

    @body.setter
    def body(self, value: str) -> None:
        self._body = value

    def __str__(self) -> str:
        return f"{self.header}\n{self.body}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name}, {self.body[:15]}...)"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Template):
            return self.name.casefold() == other.name.casefold()
        return False

    def __ne__(self, other: object) -> bool:
        return self != other

    def __hash__(self) -> int:
        return hash(self.name)

    def __lt__(self, other: Template) -> bool:
        return self.name < other.name

    def __le__(self, other: Template) -> bool:
        return self.name <= other.name

    def __gt__(self, other: Template) -> bool:
        return self.name > other.name

    def __ge__(self, other: Template) -> bool:
        return self.name >= other.name

    @classmethod
    def parse(cls, content: str) -> Template:
        """Parse a template from an ignoro string."""
        lines = content.strip().splitlines()
        pattern = re.compile(r"^#\s(\S+)\s+#$")
        headers = Template._find_metadata(lines, pattern)

        if len(headers) == 0:
            raise ignoro.exceptions.ParseError("Missing template header")
        elif len(headers) > 1:
            raise ignoro.exceptions.ParseError("Multiple template headers")

        index, name = headers[0].index, headers[0].match.group(1)
        body = cls._strip(lines[index + 2 :])

        if not body:
            raise ignoro.exceptions.ParseError(f"Missing body for '{name}'")

        return cls(name, body)

    @classmethod
    def _extract_body(cls, response: str) -> str:
        """Get the body from a gitignore.io response."""
        lines = response.splitlines()
        return cls._strip(lines[4:-1])

    @staticmethod
    def _strip(lines: Sequence[str]) -> str:
        """Strip leading and trailing whitespace from a list of lines, but preserve final newline."""
        stripped = "\n".join(lines).strip()
        return "".join(f"{line}\n" for line in stripped.splitlines())

    @staticmethod
    def _create_header(name: str, line_width: int = 100) -> str:
        """Create a block header for a template."""
        border = f"#{'-'*(line_width-2)}#"
        middle = f"# {name.upper()}{' '*(line_width-len(name)-3)}#"
        return f"{border}\n{middle}\n{border}\n"


class TemplateList(collections.abc.MutableSequence[Template], _FindMetadataMixin):
    """A list of .gitignore templates from gitignore.io."""

    def __init__(self, templates: Optional[Iterable[Template]] = None, *, populate: bool = False) -> None:
        """Initialize a list of templates from user provided templates and/or gitignore.io."""
        self.data: list[Template] = []

        if isinstance(templates, list):
            self.data = templates
        elif isinstance(templates, TemplateList):
            self.data = templates.data
        elif isinstance(templates, Iterable):
            self.data = list(templates)

        if populate:
            self.populate()

    def __str__(self) -> str:
        return "\n".join(str(template) for template in self.data)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.data!r})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TemplateList):
            return self.data == other.data
        return False

    def __ne__(self, other: object) -> bool:
        return self != other

    def __getitem__(self, index: int) -> Template:
        return self.data[index]

    def __setitem__(self, index: int, template: Template) -> None:
        self.data[index] = template

    def __delitem__(self, index: int) -> None:
        del self.data[index]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[Template]:
        return iter(self.data)

    def __contains__(self, item: object) -> bool:
        return item in self.data

    def insert(self, index: SupportsIndex, value: Template) -> None:
        """Insert a template into the list."""
        self.data.insert(index, value)

    def sort(self, key: Optional[Callable] = None, reverse: bool = False) -> None:
        """Sort the list of templates."""
        self.data.sort(key=key, reverse=reverse)

    def append(self, template: Template) -> None:
        """Append a template to the list."""
        self.data.append(template)

    def extend(self, templates: Iterable[Template]) -> None:
        """Add multiple templates to the list."""
        for template in templates:
            self.append(template)

    def replace(self, template: Template) -> None:
        """Replace a template in the list if it exists, otherwise add it."""
        if template in self.data:
            index = self.data.index(template)
            self.data[index] = template
        else:
            self.data.append(template)

    def contains(self, term: str) -> TemplateList:
        """Returns gitignore.io templates where template name contains term."""
        return TemplateList(template for template in self.data if term.casefold() in template.name.casefold())

    def startswith(self, term: str) -> TemplateList:
        """Returns gitignore.io templates where template name starts with term."""
        return TemplateList(template for template in self.data if template.name.casefold().startswith(term.casefold()))

    def match(self, term: str) -> Template | None:
        """Returns first gitignore.io template where template name matches term."""
        for template in self.data:
            if template.name.casefold() == term.casefold():
                return template

    def findall(self, terms: Iterable[str]) -> TemplateList:
        """Returns gitignore.io templates where template name matches terms."""
        terms = [term.casefold() for term in terms]
        return TemplateList(template for term in terms for template in self.data if term == template.name.casefold())

    def populate(self) -> None:
        """Populate the list of templates from gitignore.io."""
        url = f"{ignoro.BASE_URL}/list"
        params = {"format": "lines"}

        try:
            response = requests.get(url, params)
        except requests.exceptions.ConnectionError as err:
            raise ignoro.exceptions.ApiError(f"Failed to connect to '{url}'") from err
        except requests.exceptions.Timeout as err:
            raise ignoro.exceptions.ApiError(f"Connection to '{url}' timed out") from err

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise ignoro.exceptions.ApiError(f"Failed to fetch '{url}' because {err.response.reason}") from err

        template_names = response.text.splitlines()
        for name in template_names:
            self.replace(Template(name))

    @classmethod
    def parse(cls, text: str) -> TemplateList:
        """Parse templates from a string."""
        lines = text.strip().splitlines()
        pattern = re.compile(r"^#\s(\S+)\s+#$")
        headers = TemplateList._find_metadata(lines, pattern)

        if len(headers) == 0:
            raise ignoro.exceptions.ParseError("Missing template headers")

        templates = cls()
        while len(headers) > 0:
            current_header_index, _ = headers.pop(0)

            if len(headers) > 0:
                next_header_index, _ = headers[0]
                content = "".join(f"{line}\n" for line in lines[current_header_index - 1 : next_header_index - 1])
            else:
                content = "".join(f"{line}\n" for line in lines[current_header_index - 1 :])

            templates.append(Template.parse(content))

        return templates


class Gitignore(_FindMetadataMixin):
    _header: str = "# Created by https://github.com/solbero/ignoro\n"

    def __init__(self, template_list: Optional[ignoro.TemplateList] = None) -> None:
        """Initialize a .gitignore file from a list of templates."""
        self.template_list = template_list or ignoro.TemplateList()

    def __str__(self) -> str:
        return f"{self._header}\n{self.template_list}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.template_list!r})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Gitignore):
            return self.template_list == other.template_list
        return False

    def __len__(self) -> int:
        return len(self.template_list)

    def dumps(self) -> str:
        """Dump the .gitignore to a string."""
        return str(self)

    def dump(self, path: pathlib.Path) -> None:
        """Dump the .gitignore to a file."""
        try:
            if path.is_dir():
                raise IsADirectoryError(f"Path '{path.absolute()}' is a directory")
        except PermissionError as err:
            raise PermissionError(f"Permission denied for '{path.absolute()}'") from err

        try:
            path.write_text(str(self))
        except PermissionError as err:
            raise PermissionError(f"Permission denied for '{path.absolute()}'.") from err

    @classmethod
    def loads(cls, text: str) -> Gitignore:
        """Load a .gitignore from a string."""
        try:
            return cls._parse(text)
        except ignoro.exceptions.ParseError as err:
            raise ignoro.exceptions.ParseError(f"Input is invalid: {err}") from err

    @classmethod
    def load(cls, path: pathlib.Path) -> Gitignore:
        """Load a .gitignore from a file."""
        try:
            if path.is_dir():
                raise IsADirectoryError(f"Path '{path.absolute()}' is a directory")
        except PermissionError as err:
            raise PermissionError(f"Permission denied for '{path.absolute()}'.") from err

        try:
            with open(path, "r") as file:
                return cls._parse(file.read())
        except ignoro.exceptions.ParseError as err:
            raise ignoro.exceptions.ParseError(f"File '{path.absolute()}' is invalid: {err}") from err
        except FileNotFoundError as err:
            raise FileNotFoundError(f"File '{path.absolute()}' does not exist") from err
        except PermissionError as err:
            raise PermissionError(f"Permission denied for '{path.absolute()}'") from err

    @classmethod
    def _parse(cls, text: str) -> Gitignore:
        """Parse a .gitignore from a string."""
        lines = text.strip().splitlines()
        pattern = re.compile(f"^{cls._header}$")
        header = cls._find_metadata(lines, pattern)

        if header:
            content = "".join(f"{line}\n" for line in lines[header[0].index + 1 :])
        else:
            content = "".join(f"{line}\n" for line in lines)

        if not content.strip():
            return cls()

        return cls(ignoro.TemplateList.parse(content))



import functools
import pathlib

import rich
import rich.columns
import rich.console
import rich.panel
import rich.prompt
import typer
from typing_extensions import Annotated, Optional

import ignoro

__all__ = ["app"]

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
stdout = rich.console.Console(color_system="auto", highlight=False)
stderr = rich.console.Console(color_system="auto", stderr=True, highlight=False)
columns = functools.partial(rich.columns.Columns, column_first=True, equal=True, padding=(0, 2))
panel = functools.partial(rich.panel.Panel, border_style="red", title="Error", title_align="left")


@app.command("search")
def search(
    term: Annotated[
        str,
        typer.Argument(help="Term used to search [link=https://www.toptal.com/developers/gitignore]gitignore.io[/]."),
    ] = ""
):
    """
    Search for templates at [link=https://www.toptal.com/developers/gitignore]gitignore.io[/].

    If no search term is provided, all available templates will be listed.
    """
    try:
        template_list = ignoro.api.TemplateList(populate=True)
    except ignoro.exceptions.ApiError as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    template_names = tuple(template.name for template in template_list.contains(term))
    if not template_names:
        stderr.print(panel(f"No matching templates for term: '{term}'."))
        raise typer.Exit(1)

    names_underlined = tuple(name.replace(term, f"[underline]{term}[/underline]") for name in template_names)

    stdout.print(columns(names_underlined))


@app.command("create")
def create(
    templates: Annotated[
        list[str],
        typer.Argument(
            help="Templates to include in .gitignore file.",
            show_default=False,
        ),
    ],
    path: Annotated[
        Optional[pathlib.Path],
        typer.Option("--path", help="Create a .gitignore file at this path.", show_default=False),
    ] = None,
    echo: Annotated[
        bool,
        typer.Option("--show-gitignore", help="Show the result of the create command instead of writing a file."),
    ] = False,
):
    """
    Create a .gitignore file.

    If no path is provided, the .gitignore file will be created in the current directory.
    """
    if path is None:
        path = pathlib.Path.cwd() / ".gitignore"

    try:
        template_list = ignoro.api.TemplateList(populate=True)
    except ignoro.exceptions.ApiError as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    templates_matching_names = template_list.findall(templates)

    names_not_found = tuple(
        name for name in templates if name not in tuple(template.name for template in templates_matching_names)
    )
    if names_not_found:
        names_quoted = tuple(f"'{name}'" for name in names_not_found)
        stderr.print(
            panel(
                f"No matching templates for {'terms' if len(names_not_found) > 1 else 'term'}: "
                f"{', '.join(name for name in names_quoted)}.",
            )
        )
        raise typer.Exit(1)

    gitignore = ignoro.Gitignore(templates_matching_names)

    if echo:
        stdout.print(gitignore.dumps())
        raise typer.Exit(0)

    if path.exists():
        overwrite = rich.prompt.Confirm.ask(f"File {path.absolute()} already exists. Do you wish to overwrite it?")
        if not overwrite:
            raise typer.Abort()

    try:
        gitignore.dump(path)
    except (IsADirectoryError, PermissionError, ignoro.exceptions.ApiError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)


@app.command("list")
def show(
    path: Annotated[
        Optional[pathlib.Path],
        typer.Option("--path", help="List templates in .gitignore file at this path.", show_default=False),
    ] = None,
):
    """
    List templates in a .gitignore file.

    If no path is provided, the templates from the .gitignore file in the current directory will be listed.
    """
    if path is None:
        path = pathlib.Path.cwd() / ".gitignore"

    try:
        gitignore = ignoro.Gitignore.load(path)
    except (FileNotFoundError, PermissionError, IsADirectoryError, ignoro.exceptions.ParseError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    template_names = tuple(template.name for template in gitignore.template_list)

    stdout.print(columns(template_names))


@app.command("add")
def add(
    templates: Annotated[
        list[str],
        typer.Argument(
            help="Templates to add to .gitignore file.",
            show_default=False,
        ),
    ],
    path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--path",
            help="Add templates to .gitignore file at this path.",
            show_default=False,
        ),
    ] = None,
    echo: Annotated[
        bool,
        typer.Option("--show-gitignore", help="Show the result of the add command instead of writing a file."),
    ] = False,
):
    """
    Add templates to a .gitignore file.

    If no path is provided, the templates will be added to the .gitignore file in the current directory.
    """
    if path is None:
        path = pathlib.Path.cwd() / ".gitignore"

    try:
        gitignore = ignoro.Gitignore.load(path)
    except (FileNotFoundError, PermissionError, IsADirectoryError, ignoro.exceptions.ParseError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    try:
        template_list = ignoro.api.TemplateList(populate=True)
    except ignoro.exceptions.ApiError as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    templates_matching_names = template_list.findall(templates)

    names_not_found = tuple(
        name for name in templates if name not in tuple(template.name for template in templates_matching_names)
    )
    if names_not_found:
        names_quoted = tuple(f"'{name}'" for name in names_not_found)
        stderr.print(
            panel(
                f"No matching templates for {'terms' if len(names_not_found) > 1 else 'term'}: "
                f"{', '.join(name for name in names_quoted)}."
            )
        )
        raise typer.Exit(1)

    for template in templates_matching_names:
        if template in gitignore.template_list:
            overwrite = rich.prompt.Confirm.ask(
                f"Template '{template.name}' already exists in gitignore file. Do you wish to replace it?"
            )
            if overwrite:
                gitignore.template_list.replace(template)
        else:
            gitignore.template_list.append(template)

    if echo:
        stdout.print(gitignore.dumps())
        raise typer.Exit(0)

    try:
        gitignore.dump(path)
    except (IsADirectoryError, PermissionError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)


@app.command("remove")
def remove(
    templates: Annotated[
        list[str],
        typer.Argument(
            help="Templates to remove from .gitignore file.",
            show_default=False,
        ),
    ],
    path: Annotated[
        Optional[pathlib.Path],
        typer.Option("--path", help="Remove templates from .gitignore file at this path.", show_default=False),
    ] = None,
    echo: Annotated[
        bool,
        typer.Option("--show-gitignore", help="Show the result of the remove command instead of writing a file."),
    ] = False,
):
    """
    Remove templates from a .gitignore file.

    If no path is provided, the templates will be removed from the gitignore file in the current directory.
    """
    if path is None:
        path = pathlib.Path.cwd() / ".gitignore"

    try:
        gitignore = ignoro.Gitignore.load(path)
    except (FileNotFoundError, PermissionError, IsADirectoryError, ignoro.exceptions.ParseError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)

    names_not_found = tuple(
        name for name in templates if name not in tuple(template.name for template in gitignore.template_list)
    )
    if names_not_found:
        names_quoted = tuple(f"'{name}'" for name in names_not_found)
        stderr.print(
            panel(
                f"No matching templates for {'terms' if len(names_not_found) > 1 else 'term'}: "
                f"{', '.join(name for name in names_quoted)}."
            )
        )
        raise typer.Exit(1)

    templates_matching_names = gitignore.template_list.findall(templates)

    for template in templates_matching_names:
        gitignore.template_list.remove(template)

    if echo:
        stdout.print(gitignore.dumps())
        raise typer.Exit(0)

    try:
        gitignore.dump(path)
    except (IsADirectoryError, PermissionError) as err:
        stderr.print(panel(f"{err}."))
        raise typer.Exit(1)


@app.callback()
def main():
    """Create or modify a .gitignore file based on templates from [link=https://www.toptal.com/developers/gitignore]gitignore.io[/link]."""
    ...


if __name__ == "__main__":
    app()



import pathlib

import pytest
import requests
import requests_mock

import ignoro
from tests.conftest import MockErrors, TemplateMock, assert_in_string


class TestTemplate:
    def test_template_string(self, foo_template_mock: TemplateMock):
        template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)

        assert template.name == foo_template_mock.name
        assert template.body == foo_template_mock.body

    def test_template_from_remote(self, foo_template_mock: TemplateMock):
        template = ignoro.Template(foo_template_mock.name)

        assert template.name == foo_template_mock.name
        assert template.body == foo_template_mock.body

    def test_template_from_remote_error_not_found(self):
        template = ignoro.Template(MockErrors.NOT_FOUND)

        with pytest.raises(ignoro.exceptions.ApiError) as excinfo:
            template.body

        assert_in_string(("failed to fetch",), str(excinfo.value))

    def test_template_from_remote_error_connection(self):
        template = ignoro.Template(MockErrors.CONNECTION)

        with pytest.raises(ignoro.exceptions.ApiError) as excinfo:
            template.body

        assert_in_string(("failed", "connect"), str(excinfo.value))

    def test_template_from_remote_error_timeout(self):
        template = ignoro.Template(MockErrors.TIMEOUT)

        with pytest.raises(ignoro.exceptions.ApiError) as excinfo:
            template.body

        assert_in_string(("connection", "timed out"), str(excinfo.value))

    def test_template_from_local(self, foo_template_mock: TemplateMock):
        template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)

        assert template.name == foo_template_mock.name
        assert template.body == foo_template_mock.body

    def test_template_parse(self, foo_template_mock: TemplateMock):
        template = ignoro.Template.parse(foo_template_mock.content)

        assert template.name == foo_template_mock.name
        assert template.body == foo_template_mock.body

    def test_template_error_parse_empty(self):
        with pytest.raises(ignoro.exceptions.ParseError) as excinfo:
            ignoro.Template.parse("")

        assert_in_string(("missing", "header"), str(excinfo.value))

    def test_template_error_parse_missing_header(self, foo_template_mock: TemplateMock):
        with pytest.raises(ignoro.exceptions.ParseError) as excinfo:
            ignoro.Template.parse(foo_template_mock.body)

        assert_in_string(("missing", "header"), str(excinfo.value))

    def test_template_error_parse_multiple_headers(
        self, foo_template_mock: TemplateMock, bar_template_mock: TemplateMock
    ):
        with pytest.raises(ignoro.exceptions.ParseError) as excinfo:
            ignoro.Template.parse(foo_template_mock.content + "\n" + bar_template_mock.content)

        assert_in_string(("multiple", "headers"), str(excinfo.value))

    def test_template_error_parse_missing_body(self, foo_template_mock: TemplateMock):
        with pytest.raises(ignoro.exceptions.ParseError) as excinfo:
            ignoro.Template.parse(foo_template_mock.header)

        assert_in_string(("missing", "body", foo_template_mock.name), str(excinfo.value))


class TestTemplateList:
    def test_template_list(
        self,
        template_list: ignoro.TemplateList,
        template_list_names_mock: tuple[str, ...],
    ):
        template_list.populate()
        template_list_names = tuple(template.name for template in template_list)

        assert template_list_names == template_list_names_mock

    def test_template_list_str(
        self,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))

        assert str(template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

    def test_template_list_contains(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        results = tuple(template.name for template in template_list.contains("do"))

        assert results == ("dotdot", "double-dash")

    def test_templates_list_contains_no_result(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        result = template_list.contains("fizzbuzz")

        assert len(result) == 0

    def test_template_list_startswith(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        result = tuple(template.name for template in template_list.startswith("do"))

        assert result == ("dotdot", "double-dash")

    def test_template_list_startswith_no_result(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        result = tuple(template.name for template in template_list.startswith("fizzbuzz"))

        assert len(result) == 0

    def test_template_list_match(
        self,
        template_list: ignoro.TemplateList,
        foo_template_mock: TemplateMock,
    ):
        template_list.populate()
        result = template_list.match(foo_template_mock.name)

        assert result is not None
        assert result.name == foo_template_mock.name

    def test_template_list_exact_match_no_result(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        result = template_list.match("fizzbuzz")

        assert result is None

    def test_template_list_findall(
        self,
        template_list: ignoro.TemplateList,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        template_list.populate()
        result = template_list.findall((foo_template_mock.name, bar_template_mock.name))

        assert len(result) == 2
        assert result[0].name == foo_template_mock.name
        assert result[1].name == bar_template_mock.name

    def test_template_list_findall_no_result(
        self,
        template_list: ignoro.TemplateList,
    ):
        template_list.populate()
        result = template_list.findall(("fizz", "buzz"))

        assert len(result) == 0

    def test_template_list_parse(
        self,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        template_list = ignoro.TemplateList.parse(f"{foo_template_mock.content}\n{bar_template_mock.content}\n")

        assert len(template_list) == 2
        assert template_list[0].name == foo_template_mock.name
        assert template_list[0].body == foo_template_mock.body
        assert template_list[1].name == bar_template_mock.name
        assert template_list[1].body == bar_template_mock.body

    def test_templates_parse_error_no_headers(
        self,
        foo_template_mock: TemplateMock,
    ):
        with pytest.raises(ignoro.exceptions.ParseError) as excinfo:
            ignoro.TemplateList.parse(foo_template_mock.body)

        assert_in_string(
            ("missing", "template", "header"),
            str(excinfo.value),
        )

    def test_template_list_replace(
        self,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        foo_template_new_body = ignoro.Template(foo_template_mock.name, bar_template_mock.body)

        templates = ignoro.TemplateList((foo_template,))
        templates.replace(foo_template_new_body)

        assert len(templates) == 1
        assert ignoro.Template(foo_template_mock.name) in templates
        assert templates[0].body == bar_template_mock.body

    def test_template_list_extend(
        self,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        templates = ignoro.TemplateList((foo_template,))
        templates.extend(ignoro.TemplateList((bar_template,)))

        assert len(templates) == 2
        assert ignoro.Template(foo_template_mock.name) in templates
        assert ignoro.Template(bar_template_mock.name) in templates

    def test_template_list_populate_error_connection(
        self,
        requests_mock: requests_mock.Mocker,
    ):
        requests_mock.get(f"{ignoro.BASE_URL}/list?format=lines", exc=requests.exceptions.ConnectionError)

        with pytest.raises(ignoro.exceptions.ApiError) as excinfo:
            ignoro.TemplateList().populate()

        assert_in_string(("failed", "connect"), str(excinfo.value))

    def test_template_list_populate_error_timeout(
        self,
        requests_mock: requests_mock.Mocker,
    ):
        requests_mock.get(f"{ignoro.BASE_URL}/list?format=lines", exc=requests.exceptions.Timeout)

        with pytest.raises(ignoro.exceptions.ApiError) as excinfo:
            ignoro.TemplateList().populate()

        assert_in_string(("connection", "timed out"), str(excinfo.value))


class TestGitignore:
    def test_gitignore_write_and_read_string(
        self,
        template_list: ignoro.TemplateList,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))

        writer = ignoro.Gitignore(template_list)
        output = writer.dumps()
        reader = ignoro.Gitignore.loads(output)

        assert len(reader) == 2
        assert writer == reader

    def test_gitignore_write_and_read_file(
        self,
        tmp_path: pathlib.Path,
        template_list: ignoro.TemplateList,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))

        path = tmp_path / ".gitignore"
        writer = ignoro.Gitignore(template_list)
        writer.dump(path)
        reader = ignoro.Gitignore.load(path)

        assert len(reader) == 2
        assert writer == reader

    def test_gitignore_error_read_path_is_dir(
        self,
        tmp_path: pathlib.Path,
    ):
        with pytest.raises(IsADirectoryError):
            ignoro.Gitignore.load(tmp_path)

    def test_gitignore_error_read_file_not_exists(
        self,
        tmp_path: pathlib.Path,
    ):
        path = tmp_path / ".gitignore"

        with pytest.raises(FileNotFoundError):
            ignoro.Gitignore.load(path)

    def test_gitignore_error_read_permission_denied(
        self,
        tmp_path: pathlib.Path,
    ):
        path = tmp_path / ".gitignore"
        path.touch()
        path.chmod(0o000)

        with pytest.raises(PermissionError):
            ignoro.Gitignore.load(path)

    def test_gitignore_error_write_path_is_dir(
        self,
        tmp_path: pathlib.Path,
        template_list: ignoro.TemplateList,
    ):
        with pytest.raises(IsADirectoryError):
            ignoro.Gitignore(template_list).dump(tmp_path)

    def test_gitignore_error_write_permission_denied(
        self,
        tmp_path: pathlib.Path,
        template_list: ignoro.TemplateList,
    ):
        path = tmp_path / ".gitignore"
        path.touch()
        path.chmod(0o000)

        with pytest.raises(PermissionError):
            ignoro.Gitignore(template_list).dump(path)



import pytest
import requests
import requests_mock
from conftest import TemplateMock, TestConsole, assert_in_string

import ignoro.cli
from ignoro.api import Gitignore


class TestSearchCommand:
    def test_search_all(
        self,
        test_console: TestConsole,
        template_list_names_mock: tuple[str, ...],
    ):
        result = test_console.runner.invoke(ignoro.app, ("search",))

        assert result.exit_code == 0
        assert tuple(str.split(result.stdout)) == template_list_names_mock

    def test_search_term(
        self,
        test_console: TestConsole,
    ):
        name = "do"
        result = test_console.runner.invoke(ignoro.app, ("search", name))

        assert result.exit_code == 0
        assert tuple(str.split(result.stdout)) == ("dotdot", "double-dash")

    def test_search_search_no_result(
        self,
        test_console: TestConsole,
    ):
        name = "fizzbuzz"
        result = test_console.runner.invoke(ignoro.app, ("search", name))

        assert result.exit_code == 1
        assert_in_string(("error", "no matching", name), result.stderr)

    @pytest.mark.parametrize(
        "error, fragments",
        (
            (requests.exceptions.Timeout, ("error", "connection", "timed", "out")),
            (requests.exceptions.ConnectionError, ("error", "failed", "connect")),
        ),
    )
    def test_search_error_remote_list_not_found(
        self,
        test_console: TestConsole,
        requests_mock: requests_mock.Mocker,
        error: requests.exceptions.RequestException,
        fragments: tuple[str, ...],
    ):
        requests_mock.get(f"{ignoro.BASE_URL}/list?format=lines", exc=error)
        result = test_console.runner.invoke(ignoro.app, ("search",))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(fragments, result.stderr)


class TestCreateCommand:
    def test_create_file(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        result = test_console.runner.invoke(ignoro.app, ("create", foo_template_mock.name))

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == foo_template_mock.content

    def test_create_file_at_path(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        subdir = test_console.cwd / "subdir"
        subdir.mkdir()
        path = subdir / ".gitignore"

        result = test_console.runner.invoke(
            ignoro.app, ("create", foo_template_mock.name, "--path", str(path)), terminal_width=100
        )

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == foo_template_mock.content

    @pytest.mark.xfail(reason="Setting terminal width does not change the width of the terminal in the test.")
    def test_create_show(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(
            ignoro.app, ("create", foo_template_mock.name, "--show-gitignore"), terminal_width=100
        )

        assert result.exit_code == 0
        assert str(Gitignore.loads(result.stdout).template_list) == foo_template_mock.content

    def test_create_file_two_templates(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        result = test_console.runner.invoke(ignoro.app, ("create", foo_template_mock.name, bar_template_mock.name))

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

    def test_create_file_already_exists_overwrite(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        path.touch()
        result = test_console.runner.invoke(ignoro.app, ("create", foo_template_mock.name), input="y\n")

        assert result.exit_code == 0
        assert_in_string(("already exists", "overwrite"), result.stdout)

    def test_create_file_already_exists_abort(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        path.touch()
        result = test_console.runner.invoke(ignoro.app, ("create", foo_template_mock.name), input="n\n")

        assert result.exit_code == 1
        assert_in_string(("already exists", "overwrite"), result.stdout)
        assert_in_string(("aborted",), result.stderr)

    @pytest.mark.parametrize(
        "error, fragments",
        (
            (requests.exceptions.ConnectionError, ("error", "failed", "connect")),
            (requests.exceptions.Timeout, ("error", "connection", "timed", "out")),
        ),
    )
    def test_create_error_remote_template_not_found(
        self,
        test_console: TestConsole,
        requests_mock: requests_mock.Mocker,
        foo_template_mock: TemplateMock,
        error: requests.exceptions.RequestException,
        fragments: tuple[str, ...],
    ):
        requests_mock.get(f"{ignoro.BASE_URL}/{foo_template_mock.name}", exc=error)
        result = test_console.runner.invoke(ignoro.app, ("create", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(fragments, result.stderr)

    def test_create_error_template_not_found(
        self,
        test_console: TestConsole,
    ):
        name = "fizzbuzz"
        result = test_console.runner.invoke(ignoro.app, ("create", name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "no matching", name), result.stderr)

    def test_create_error_path_is_dir(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(
            ignoro.app, ("create", foo_template_mock.name, "--path", str(test_console.cwd)), input="y\n"
        )

        assert result.exit_code == 1
        assert_in_string(("error", "directory"), result.stderr)

    def test_create_error_permission_denied(
        self,
        test_console: TestConsole,
    ):
        test_console.cwd.chmod(0o0555)
        result = test_console.runner.invoke(ignoro.app, ("create", "foo"))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "permission denied"), result.stderr)


class TestListCommand:
    def test_list(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("list",))

        assert result.exit_code == 0
        assert tuple(result.stdout.split()) == (foo_template_mock.name, bar_template_mock.name)

    def test_list_at_path(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        subdir = test_console.cwd / "subdir"
        subdir.mkdir()
        path = subdir / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("list", "--path", f"{str(path)}"))

        assert result.exit_code == 0
        assert tuple(result.stdout.split()) == (foo_template_mock.name, bar_template_mock.name)

    def test_list_error_file_not_exists(
        self,
        test_console: TestConsole,
    ):
        result = test_console.runner.invoke(ignoro.app, ("list",))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "file", "not exist"), result.stderr)

    def test_list_error_permission_denied(
        self,
        test_console: TestConsole,
    ):
        path = test_console.cwd / ".gitignore"
        path.touch()
        path.chmod(0o000)

        result = test_console.runner.invoke(ignoro.app, ("list",))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "permission denied"), result.stderr)

    def test_list_error_file_invalid(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        path.write_text(str(foo_template_mock.body))

        result = test_console.runner.invoke(ignoro.app, ("list",))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "file", "invalid", "missing", "header"), result.stderr)

    def test_list_error_empty_body(
        self,
        test_console: TestConsole,
    ):
        template_list = ignoro.TemplateList()
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("list",))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "missing", "body"), result.stderr)


class TestAddCommand:
    def test_add(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("add", bar_template_mock.name))

        assert result.exit_code == 0
        assert len(Gitignore.load(path).template_list) == 2
        assert str(Gitignore.load(path).template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

    def test_add_at_path(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_temlate = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        template_list = ignoro.TemplateList((foo_temlate,))
        gitignore = ignoro.Gitignore(template_list)

        subdir = test_console.cwd / "subdir"
        subdir.mkdir()
        path = subdir / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(
            ignoro.app, ("add", bar_template_mock.name, "--path", str(path)), terminal_width=100
        )

        assert result.exit_code == 0
        assert len(Gitignore.load(path).template_list) == 2
        assert str(Gitignore.load(path).template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

    @pytest.mark.xfail(reason="Setting terminal width does not change the width of the terminal in the test.")
    def test_add_show(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(
            ignoro.app, ("add", bar_template_mock.name, "--show-gitignore"), terminal_width=100
        )
        gitignore = Gitignore.loads(result.stdout)

        assert result.exit_code == 0
        assert len(gitignore) == 2
        assert str(gitignore.template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

    def test_add_to_existing_confirmed(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("add", foo_template_mock.name), input="y\n")

        assert result.exit_code == 0
        assert_in_string(("already exists", "replace"), result.stdout)
        assert len(Gitignore.load(path).template_list) == 1
        assert str(Gitignore.load(path).template_list) == foo_template_mock.content

    def test_add_to_existing_declined(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("add", foo_template_mock.name), input="n\n")

        assert result.exit_code == 0
        assert_in_string(("already exists", "replace"), result.stdout)
        assert len(Gitignore.load(path).template_list) == 1
        assert str(Gitignore.load(path).template_list) == str(template_list)

    def test_add_error_template_not_found(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        name = "fizzbuzz"
        result = test_console.runner.invoke(ignoro.app, ("add", name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("no matching", name), result.stderr)

    def test_add_error_file_invalid(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        path.write_text(str(foo_template_mock.body))

        result = test_console.runner.invoke(ignoro.app, ("add", bar_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "file", "invalid", "missing", "header"), result.stderr)

    def test_add_error_file_not_exists(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(ignoro.app, ("add", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "not exist"), result.stderr)

    def test_add_error_path_is_dir(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(
            ignoro.app, ("add", foo_template_mock.name, "--path", str(test_console.cwd))
        )

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "directory"), result.stderr)

    def test_add_error_permission_denied(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        template_list = ignoro.TemplateList((foo_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)
        path.chmod(0o0555)

        result = test_console.runner.invoke(ignoro.app, ("add", bar_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("permission denied"), result.stderr)


class TestRemove:
    def test_remove(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("remove", foo_template_mock.name))

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == bar_template_mock.content

    def test_remove_at_path(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        subdir = test_console.cwd / "subdir"
        subdir.mkdir()
        path = subdir / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(
            ignoro.app, ("remove", foo_template_mock.name, "--path", str(path)), terminal_width=100
        )

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == bar_template_mock.content

    @pytest.mark.xfail(reason="Setting terminal width does not change the width of the terminal in the test.")
    def test_remove_show(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)
        result = test_console.runner.invoke(
            ignoro.app, ("remove", foo_template_mock.name, "--show-gitignore"), terminal_width=100
        )

        assert result.exit_code == 0
        assert str(Gitignore.loads(result.stdout).template_list) == bar_template_mock.content

    def test_remove_error_path_is_dir(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(
            ignoro.app, ("remove", foo_template_mock.name, "--path", str(test_console.cwd))
        )

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "directory"), result.stderr)

    def test_remove_error_file_invalid(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"
        path.write_text(foo_template_mock.body)

        result = test_console.runner.invoke(ignoro.app, ("remove", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "file", "invalid", "missing", "header"), result.stderr)

    def test_remove_error_template_not_found(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((bar_template,))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)

        result = test_console.runner.invoke(ignoro.app, ("remove", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "no matching", foo_template_mock.name), result.stderr)

    def test_remove_error_file_not_exists(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        result = test_console.runner.invoke(ignoro.app, ("remove", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "not exist"), result.stderr)

    def test_remove_error_write_permission_denied(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        foo_template = ignoro.Template(foo_template_mock.name, foo_template_mock.body)
        bar_template = ignoro.Template(bar_template_mock.name, bar_template_mock.body)
        template_list = ignoro.TemplateList((foo_template, bar_template))
        gitignore = ignoro.Gitignore(template_list)

        path = test_console.cwd / ".gitignore"
        gitignore.dump(path)
        path.chmod(0o0555)

        result = test_console.runner.invoke(ignoro.app, ("remove", foo_template_mock.name))

        assert result.exit_code == 1
        assert result.stdout == ""
        assert_in_string(("error", "permission denied"), result.stderr)


class TestIntegration:
    def test_search_and_create(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"

        result = test_console.runner.invoke(ignoro.app, ["search", foo_template_mock.name])

        assert result.exit_code == 0
        assert result.stdout.split() == ["foo", "foobar"]

        result = test_console.runner.invoke(ignoro.app, ["create", foo_template_mock.name])

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == foo_template_mock.content

    def test_create_add_remove_and_list(
        self,
        test_console: TestConsole,
        foo_template_mock: TemplateMock,
        bar_template_mock: TemplateMock,
    ):
        path = test_console.cwd / ".gitignore"

        result = test_console.runner.invoke(ignoro.app, ["create", foo_template_mock.name])

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == foo_template_mock.content

        result = test_console.runner.invoke(ignoro.app, ["add", bar_template_mock.name])

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == f"{foo_template_mock.content}\n{bar_template_mock.content}"

        result = test_console.runner.invoke(ignoro.app, ["remove", foo_template_mock.name])

        assert result.exit_code == 0
        assert str(Gitignore.load(path).template_list) == bar_template_mock.content

        result = test_console.runner.invoke(ignoro.app, ["list"])

        assert result.exit_code == 0
        assert result.stdout.split() == [bar_template_mock.name]
